/* Copyright (c) 2005 Miller Puckette.  The contents of this file are free for
any use, but THE AUTHOR DISCLAIMS ALL WARRANTIES related to it. */

#include "m_pd.h"
#include <math.h>
#include <stdio.h>
#include <alloca.h>
#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif

static t_class *npeak_class;
#define NEGBINS 4

typedef struct _npeak
{
    t_object x_obj;
    float x_param1;
    float x_param2;
    float x_param3;
} t_npeak;

typedef struct peak
{
    float p_pit;
    float p_amp;
    float p_real;
    float p_imag;
} t_peak;

static void *npeak_new( void)
{
    t_npeak *x = (t_npeak *)pd_new(npeak_class);
    x->x_param1 = 0;
    x->x_param2 = 0;
    x->x_param3 = 0;

    outlet_new(&x->x_obj, &s_list);
    return (x);
}

/* parameters for von Hann window (change these to get Hamming if desired) */
#define W_ALPHA 0.5
#define W_BETA 0.5

static float sinx(float theta, float sintheta)
{
    if (theta > -0.003 && theta < 0.003)
        return(1);
    else return (sintheta/theta);
}

#define PI 3.14159265

static float window_altmag(float pidetune, float sinpidetune)
{
    return (W_ALPHA * sinx(pidetune, sinpidetune)
        - 0.5 * W_BETA *
            (sinx(pidetune+PI, sinpidetune) + sinx(pidetune-PI, sinpidetune)));
}

static float window_mag(float pidetune, float cospidetune)
{
    return (sinx(pidetune + (PI/2), cospidetune)
        + sinx(pidetune - (PI/2), -cospidetune));
}

static void npeak_getrawpeaks(int npts, float *insamps, int npeak,
    int *nfound, float *fpfreq, float *fpamp,
    	float *fpampre, float *fpampim,
        float param1, float param2, float param3, float srate, int loud)
{
    float oneovern = 1.0/ (float)npts;
    float fperbin = 0.5 * srate * oneovern;
    int npts2 = 2*npts, i, bin;
    int count, peakcount = 0;
    float *fp1, *fp2;
    float *rawpow, *rawreal, *rawimag, *maskbuf;
    float *bigbuf = alloca(sizeof (float ) * (2*NEGBINS + 5*npts));
    int altwind = (param3 > 0);
    maskbuf = bigbuf + npts2;
    rawreal = maskbuf + npts+NEGBINS;
    rawimag = rawreal+npts+NEGBINS;

    for (i = 0; i < npts; i++)
        maskbuf[i] = 0;
    for (i = 0; i < npts; i++)
        bigbuf[i] = insamps[i];
    for (i = npts; i < 2*npts; i++)
        bigbuf[i] = 0;
    mayer_realfft(npts2, bigbuf);
    for (i = 0; i < npts; i++)
        rawreal[i] = bigbuf[i];
    rawreal[-1] = rawreal[1];
    rawreal[-2] = rawreal[2];
    rawreal[-3] = rawreal[3];
    rawreal[-4] = rawreal[4];
    for (i = 1; i < npts-1; i++)
        rawimag[i] = bigbuf[npts2-i];
    rawimag[0] = rawimag[npts-1] = 0;
    rawimag[-1] = -rawimag[1];
    rawimag[-2] = -rawimag[2];
    rawimag[-3] = -rawimag[3];
    rawimag[-4] = -rawimag[4];

    for (peakcount = 0; peakcount < npeak; peakcount++)
    {
        float maxpower = 0, totalpower = 0, windreal, windimag, windpower,
            detune, pidetune, sinpidetune, cospidetune, ampcorrect, ampout,
            ampoutreal, ampoutimag, freqout, freqcount1, freqcount2, powmask;
        int bestindex = -1;
        for (bin = 2, fp1 = rawreal+2, fp2 = rawimag+2;
            bin < npts - NEGBINS; bin++, fp1++, fp2++)
        {
            float x1, x2, a1, a2, b1, b2, thresh;
            if (altwind)
            {
	        x1 = fp1[0] - 0.5*(fp1[-2] +fp1[2]);
	        x2 = fp2[0] - 0.5*(fp2[-2] +fp2[2]);
	        a1 = fp1[4] - 0.5*(fp1[2] +fp1[6]);
	        a2 = fp2[2] - 0.5*(fp2[2] +fp2[6]);
	        b1 = fp1[-4] - 0.5*(fp1[-2] +fp1[-6]);
	        b2 = fp2[-4] - 0.5*(fp2[-2] +fp2[-6]);
                thresh = param2 * (a1*a1+a2*a2+b1*b1+b2*b2);
            }
            else
            {
	        x1 = fp1[1] - fp1[-1];
	        x2 = fp2[1] - fp2[-1];
	        a1 = fp1[3] - fp1[1];
	        a2 = fp2[3] - fp2[1];
	        b1 = fp1[-1] - fp1[-3];
	        b2 = fp2[-1] - fp2[-3];
                thresh = param2 * (a1*a1+a2*a2+b1*b1+b2*b2);
            }
            float pow1 = x1*x1+x2*x2;
            if (pow1 > maxpower && pow1 > maskbuf[bin])
            {
                if (pow1 > thresh)
                    maxpower = pow1, bestindex = bin;
            }
            totalpower += pow1;
        }
        if (totalpower <= 0 || maxpower < 1e-10*totalpower || bestindex < 0)
            break;
        fp1 = rawreal+bestindex;
        fp2 = rawimag+bestindex;
	
	powmask = maxpower * exp(-param1 * log(10.) / 10.);
        if (loud > 1)
            post("maxpower %f, powmask %f, param1 %f",
                maxpower, powmask, param1);
        for (bin = 2; bin < npts - NEGBINS; bin++)
        {
	    float bindiff = bin - bestindex;
            float mymask =
                powmask/ (1. + bindiff * bindiff * bindiff * bindiff);
            if (bindiff < 2 && bindiff > -2)
                mymask = 2*maxpower;
	    if (mymask > maskbuf[bin])
	    	maskbuf[bin] = mymask;
	} 
        
        if (loud)
	    post("best index %d, total power %f", bestindex, totalpower);
        if (altwind)
        {
    	    windreal = 0.5 * fp1[0] - 0.25 * (fp1[2] + fp1[-2]);
            windimag = 0.5 * fp2[0] - 0.25 * (fp2[2] + fp2[-2]);
            windpower = windreal * windreal + windimag * windimag;
            detune = (
                (W_BETA*(rawreal[bestindex-2] - rawreal[bestindex+2])) *
    	        (2.0 * W_ALPHA * rawreal[bestindex] - 
                    W_BETA * (rawreal[bestindex-2] + rawreal[bestindex+2]))
                        +
                (W_BETA*(rawimag[bestindex-2] - rawimag[bestindex+2])) *
    	        (2.0 * W_ALPHA * rawimag[bestindex] - 
                    W_BETA * (rawimag[bestindex-2] + rawimag[bestindex+2]))
    	    	    	    ) / (4.0 * windpower);
        }
        else
        {
    	    windreal = fp1[1] - fp1[-1];
            windimag = fp2[1] - fp2[-1];
            windpower = windreal * windreal + windimag * windimag;
            detune = ((fp1[1] * fp1[1] - fp1[-1]*fp1[-1]) 
                + (fp2[1] * fp2[1] - fp2[-1]*fp2[-1])) / (2 * windpower);
            if (loud > 1) post("(-1) %f %f; (1) %f %f",
                fp1[-1], fp2[-1], fp1[1], fp2[1]);
            if (loud > 1) post("peak %f %f",
                fp1[0], fp2[0]);
        }
        if (detune > 0.5)
            detune = 0.5;
        else if (detune < -0.5)
            detune = -0.5;
        if (loud > 1)
            post("windpower %f, index %d, detune %f",
                windpower, bestindex, detune);
    	pidetune = PI * detune;
    	sinpidetune = sin(pidetune);
    	cospidetune = cos(pidetune);
    	if (altwind)
	    ampcorrect = 1.0 / window_altmag(pidetune, sinpidetune);
    	else ampcorrect = 1.0 / window_mag(pidetune, cospidetune);

    	ampout = ampcorrect *sqrt(windpower);
    	ampoutreal = ampcorrect *
    	    (windreal * cospidetune - windimag * sinpidetune);
    	ampoutimag = ampcorrect *
    	    (windreal * sinpidetune + windimag * cospidetune);

    	    /* the frequency is the sum of the bin frequency and detuning */

    	*fpfreq++ = (freqout = (bestindex + 2*detune)) * fperbin;
    	*fpamp++ = oneovern * ampout;
    	*fpampre++ = oneovern * ampoutreal;
    	*fpampim++ = oneovern * ampoutimag;
#if 0
            /* and now go through and subtract the sinh function from 
            all the bins of the FFT */

        for (bin = 0, freqcount1 = -(PI*0.5)*freqout, freqcount2 = (PI*0.5)*freqout,
	    fp1 = rawreal, fp2 = rawimag;
            	bin < npts; bin++, fp1++, fp2++,
		    freqcount1 += (PI*0.5), freqcount2 += (PI*0.5))
        {
	    float subre2, subim2;
            float sx = sinx(freqcount1, sin(freqcount1));
	    float phasere = cos(freqcount1 * (npts-0.5)*oneovern);
	    float phaseim = sin(freqcount1 * (npts-0.5)*oneovern);
	    float subre1 = sx * (phasere * ampoutreal - phaseim * ampoutimag);
	    float subim1 = sx * (phaseim * ampoutreal + phasere * ampoutimag);

            sx = sinx(freqcount2, sin(freqcount2));
	    phasere = cos(freqcount2 * (npts-0.5)*oneovern);
	    phaseim = sin(freqcount2 * (npts-0.5)*oneovern);
	    subre2 = sx * (phasere * ampoutreal + phaseim * ampoutimag);
	    subim2 = sx * (phaseim * ampoutreal - phasere * ampoutimag);

    	    if ((loud > 1) && (bin < 20))
	    	post("sub %f, %f  orig %f, %f",
		    subre1, subim1, *fp1, *fp2);
            *fp1 -= (subre1 + subre2);
            *fp2 -= (subim1 + subim2);
        }
#endif
    }
    *nfound = peakcount;
}

static void npeak_list(t_npeak *x, t_symbol *s, int argc, t_atom *argv)
{
    int npts = atom_getintarg(0, argc, argv);
    t_symbol *syminput = atom_getsymbolarg(1, argc, argv);
    int onset = atom_getintarg(2, argc, argv);
    int npeak = atom_getintarg(3, argc, argv);
    int loud = atom_getintarg(4, argc, argv);
    float srate = atom_getfloatarg(5, argc, argv);
    t_float *bigbuf, *fpfreq, *fpamp, *fpampre, *fpampim;
    int arraysize, totstorage, nfound, i;
    t_garray *a;
    float *arraypoints;
    
    if (argc < 6)
    {
        post(
         "npeak: npts, array-name, array-onset, npeaks, loud, samplerate");
        return;
    }
    if (npts < 64 || npeak < 1 || npts != (1 << ilog2(npts))) 
    {
        error("npeak: bad npoints or npeak");
        return;
    }
    bigbuf = (float *)alloca(4 * sizeof(float) * npeak);
    if (onset < 0)
    {
        error("npeak: negative onset");
        return;
    }
    
    if (!(a = (t_garray *)pd_findbyclass(syminput, garray_class)) ||
    	!garray_getfloatarray(a, &arraysize, &arraypoints) ||
    	    arraysize < onset + npts)
    	    	error("%s: array missing or too small", syminput->s_name);

    fpfreq = bigbuf;
    fpamp = bigbuf + npeak;
    fpampre = bigbuf + 2 * npeak;
    fpampim = bigbuf + 3 * npeak;

    npeak_getrawpeaks(npts, arraypoints, npeak,
    	&nfound, fpfreq, fpamp, fpampre, fpampim,
            x->x_param1, x->x_param2, x->x_param3, srate, loud);
    for (i = 0; i < nfound; i++, fpamp++, fpfreq++, fpampre++, fpampim++)
    {
    	t_atom at[5];
    	SETFLOAT(at, (float)i);
    	SETFLOAT(at+1, *fpfreq);
    	SETFLOAT(at+2, *fpamp);
    	SETFLOAT(at+3, *fpampre);
    	SETFLOAT(at+4, *fpampim);
    	outlet_list(x->x_obj.ob_outlet, &s_list, 5, at);   
    }
}

static void npeak_param1(t_npeak *x, t_floatarg f)
{
    x->x_param1 = f;
}

static void npeak_param2(t_npeak *x, t_floatarg f)
{
    x->x_param2 = f;
}

static void npeak_param3(t_npeak *x, t_floatarg f)
{
    x->x_param3 = f;
}

static void npeak_free(t_npeak *x)
{
}

void npeak_setup(void)
{
    npeak_class = class_new(gensym("npeak"), (t_newmethod)npeak_new,
        (t_method)npeak_free, sizeof(t_npeak),0, 0);
    class_addlist(npeak_class, npeak_list);
    class_addmethod(npeak_class, (t_method)npeak_param1,
    	gensym("param1"), A_FLOAT, 0);
    class_addmethod(npeak_class, (t_method)npeak_param2,
    	gensym("param2"), A_FLOAT, 0);
    class_addmethod(npeak_class, (t_method)npeak_param3,
    	gensym("param3"), A_FLOAT, 0);
    post("npeak v. 0.01");
}

